package btree.visitor;
import btree.*;
import java.util.*;

/**
 * Algorithm to insert a given new data element into a 
 * search tree as per the search criteria.
 * The ordering is determined by a Comparator object
 */
public class InsertSearchTree implements IBTreeAlgo {
  /**
   * The order (max # of data elements) of the tree
   */
  private int _order;
  
  /**
   * The factory used to instantiate new B-Trees when necessary
   */
  private IBTreeFac _fac;
  
  /**
   * The Comparator object used to determine the ordering of data 
   * elements.  By default is compares two Comparable objects.
   */
  private Comparator _comp = new Comparator() {
    public int compare(Object x, Object y) {
      return ((Comparable)x).compareTo(y);
    }
  };
  
  /**
   * Convenience value that references an empty tree
   * This value is initialize in the constructor using the 
   * IBTreeFac instance.
   */
  private IMTBTree _mt;
  
  /**
   * Constructor for the class that uses the default Comparator instance.
   * @param order The order of the B-Tree that is created.
   * @param fac The factory used to create the new trees.
   */
  public InsertSearchTree(int order, IBTreeFac fac){
    _order = order;
    _fac = fac;
    _mt = fac.makeMTBTree();
  }
  
  /**
   * Constructor for the class that uses the given Comparator instance.
   * @param order The order of the B-Tree that is created.
   * @param fac The factory used to create the new trees.
   * @param comp The Comparator object to use to determine ordering.
   */
  public InsertSearchTree(int order, IBTreeFac fac, Comparator comp){
    this(order, fac);
    _comp = comp;
  }
  
  /**
   * Tree is empty.  Return a new INEBTree that has exactly one data 
   * element, params[0], and whose children are all empty.
   * @param params params[0] is the new data element to store.
   * @return A new INEBTree instance.
   */
  public Object mtCase(IMTBTree host, Object...params) {
    
    return _fac.makeNEBTree(new Object[]{params[0]},
                            new IBTree[]{_mt, _mt});
  }
  
  /**
   * Tree is non-empty. 
   * If the number of root data elements is less than the order 
   * and the child tree at the insertion point is empty, then insert
   * the new data element into this node's data and add an empty 
   * child tree.  
   * If the number of root data elements is less than the order 
   * and the child tree at the insertion point is NOT empty,
   * insert the data element into the child tree.
   * If the number of root data elements is equal to 
   * the order, then always insert into the child tree at 
   * the insertion point.
   * A NEW INEBTree is ALWAYS returned because the tree is
   * immutable.
   * @param params params[0] is the new data element to store.
   * @return A new INEBTree instance.
   */
  public Object neCase(INEBTree host, final Object...params){
    final Object[] data = host.getData(); // gets a copy
    final IBTree[] children = host.getChildren(); // gets a copy
    final int idx = getInsertIndex(data, params[0]);
    
// recur only if child is non-empty and not at max size.
    if(data.length < _order) {
      // not full, depends on whether or not child is empty
      // delegate to appropriate child
      return children[idx].execute(new IBTreeAlgo() {
        /**
         * Child is empty, so can insert into host
         */
        public Object mtCase(IMTBTree childHost, Object...nu) {
          Object[] newData = new Object[data.length+1];
          IBTree[] newChildren = new IBTree[data.length+2];
          newData[idx] = params[0];
          newChildren[idx] = _mt;
          
          for(int i=0, j=0; i<data.length; i++, j++) {
            if(i == idx)  j++; // skip over new data
            newData[j] = data[i];
            newChildren[j] = children[i];
          }
          newChildren[newData.length] = children[data.length];
          
          // insert new data elements with new children
          return _fac.makeNEBTree(newData, newChildren);
        }
        
        /**
         * Child is non-empty, so must insert into child
         */
        public Object neCase(INEBTree childHost, Object...nu) {
          // childHost and children[idx] refer to the same tree
          children[idx] = (IBTree) childHost.execute(InsertSearchTree.this, params);
          return _fac.makeNEBTree(data, children);
        }
        
      });
    }
    else {  // This level is already at max capacity, so, recur to child
      children[idx] = (IBTree) children[idx].execute(this, params);
      return _fac.makeNEBTree(data, children);
    }
  }
  
  /**
   * Finds the index of the child tree where a new data element 
   * would be inserted with respect to a given data array, which 
   * is assumed to be sorted in ascending order as per the _comp object.
   * 0 returned means that x is smaller than the first element in array.
   * array.length returned means that x is larger than the last element in array.
   */
  public int getInsertIndex(Object[] array, Object x) {

    for(int i=0;i<array.length;i++) {
      if(0 < _comp.compare(array[i], x)){
        return i;
      }
    }
    return array.length;
  }
}