package life.model;

/**
 * Abstract evolution algorithm that provides utility objects 
 * and methods for its subclasses.
 * Any non-overridden case methods in the subclass will be no-ops
 * that return null.
 */
public abstract class AEvolveAlgo extends ACellAlgo {
  
  /**
   * Factory used to create lambdas that will change the state of 
   * a cell when applied.
   */
  protected CellStateChangeFactory fac = CellStateChangeFactory.Singleton;
  
  /**
   * The ICellAlgo that the countNeighbors method executes on every cell 
   * it processes.
   */
  protected ICellAlgo countAlgo = new ACellAlgo() {};
  
  /**
   * Counts the number of neighbors of different types over the sets
   * [row-range, row+range] and [col-range, col+range]  (inclusive).
   * @param env  The LifeModel environment being used.
   * @param row The row of the cell whose neighbors are being counted.
   * @param col The col of the cell whose neighbors are being counted.
   * @param range The distance away from the cell that will be considered neighbors, e.g. range =1 will count only the nearest neighbors.
   * @return An array of integers with the counts for each state of an ICell.
   */
  protected int[] countNeighbors(LifeModel env, int row, int col, int range) {
    int[] totals = new int[ICell.CELLSTATES.values().length];
//    for(int i = row-range; i<= row+range; i++) {
//      for(int j = col-range; j <= col+range; j++) {
//        if(row != i || col != j) {
//          env.getCellAt(i, j).execute(countAlgo, totals);
//        }
//      }
//    }
    for(int i = -range; i<= range; i++) {
      for(int j = -range; j <= +range; j++) {
        if(0 != i || 0 != j) {
          env.getCellAt(row+i, col+j).execute(countAlgo, totals);
        }
      }
    }
    return totals;
  }
  
}