package life.model;
import javax.swing.Timer;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Color;
import lrs.*;
import util.*;
import java.lang.reflect.*;

/**
 * Main model of the system.  This is the environment in which the cells "live".
 */
public class LifeModel {
  /**
   * numbers of rows and columns of cells in the system
   */
  private static int NCELLROWS = 32;
  private static int NCELLCOLS = 32;
  
  /**
   * Delay between updates to the system
   */
  private static int DELAY = 300; // milliseconds
  
  /**
   * The adapter to the view
   */
  private IViewAdapter view;
 
  /** 
   * IAlgo used to run all the lambdas that change the states
   * of the cells.  The lambdas are stored in an LRStruct
   */
  private IAlgo runLambdas = new IAlgo() {
    public Object emptyCase(LRStruct host, Object... nu) {
      return null;
    }
    public Object nonEmptyCase(LRStruct host, Object... nu) {
      ((ILambda)host.getFirst()).apply();
      return host.getRest().execute(this);
    }
  };
  
  /**
   * The current evolution strategy to use
   */
  private ICellAlgo strategy = Evolve1Algo.Singleton;
  
  /**
   * Timer that causes a periodic updating of the system
   */
  private Timer timer = new Timer(DELAY, new ActionListener() {
    public void actionPerformed(ActionEvent e) {
      // First, calculate what cells will do, saving that as a lambda
      final LRStruct lambdas = new LRStruct();
      

      // Prob. 2.a.ii: STUDENT TO REPLACE THE FOLLOW FOR-LOOP WITH AN ILambda
      // AND A CALL TO mapRowCol().
      // The ILambd's apply should be given 3 input parameters:
      // The cell, its row and its column in that order.
      // Note, ILambda.apply() takes Objects as inputs. 
      // The lambda should return null.
      
      for(int row = 0; row<NCELLROWS; row++) {
        for(int col = 0; col<NCELLCOLS; col++) {
          cells[row][col].execute(strategy, row, col, lambdas);
        }
      }
      


      // Second, run all the lambdas.
      lambdas.execute(runLambdas);
      // Third, update the view with the new cell states.
      view.update();
    }
  });
  
  
  /**
   * The 2-dimensional array of cells in the system
   * A cell is accessed as cells[row][col]
   */
  private ICell[][] cells;
  
  /**
   * Constructor for the class
   * @param view The adapter to the view
   */
  public LifeModel(IViewAdapter view) {
    this.view = view;
    // Initialize the cells
    // All cells start dead
    cells = new ICell[NCELLROWS][NCELLCOLS];
    for(int row = 0; row<NCELLROWS; row++) {
      for(int col = 0; col<NCELLCOLS; col++) {
        cells[row][col] = new Cell(this);
      }
    }
  }
  
  /**
   * GEt number of rows of cells in the system.
   */
  public int getNRows() {
    return NCELLROWS;
  }
  
  /**
   * GEt the number of colums of cells in the system
   */
  public int getNCols() {
    return NCELLCOLS;
  }
  
  /**
   * STart the model up.
   * starts the periodic updating of the cells.
   */
  public void start() {
    timer.start();
  }
  
  /**
   * Stops the model by stopping the periodic updating of the cells.
   */
  public void stop() {
    timer.stop();
  }
  
  //Prob. 2.b.ii STUDENT TO WRITE LAMBDA TO BE USED IN RESET() METHOD
  private ILambda killCell = null;  //STUB CODE
  
  /**
   * Resets all the cells to be dead
   */
  public void reset(){
    
    // Prob. 2.b.ii: STUDENT TO REPLACE THE FOLLOWING DOUBLE LOOP WITH A CALL TO 
    // MAP USING THE ABOVE killCell ILAMBDA.
    
    for(int row = 0; row<NCELLROWS; row++) {
      for(int col = 0; col<NCELLCOLS; col++) {
        cells[row][col].setTo(ICell.CELLSTATES.DEAD);
      }
    }
    
    view.update();
  }
  
  /**
   * GEt the color of the cell at the given row and col
   */
  public Color getColorAt(int row, int col) {
    return cells[row][col].getColor();
  }
  
  /**
   * Get the name of the current evolution strategy in use.
   */
  public String getStrategyName() {
    return strategy.getClass().getName();
  }
  
  /**
   * Dynamically load the given Singleton ICellAlgo visitor
   * given its fully qualified class name.
   * This method does nothing if an exception is encountered.
   * @param s The fully qualified classname of the desired ICellAlgo
   * @return The fully qualified classname of the value of strategy after this method is done.
   */
  public String setStrategyByName(String s) {
    try {
      Class c = Class.forName(s);
      Field f = c.getField("Singleton");
      strategy = (ICellAlgo) f.get(null);
    }
    catch (Exception e) {
      System.out.println("No such class or is not a Singleton: "+ s);
    }
    return getStrategyName();
  }
  
  
  /**
   * change the given cell to its next possible state.
   */
  public Color toNextState(int row, int col) {
    return cells[row][col].toNextState();
  }
  
  /**
   * Get the cell at the given row and col.
   * We have a toroidal universe which means that 
   * the ends wrap around, e.g. if you keep going to 
   * the right, you will reappear on the left.
   * System will always retrieve a cell, no matter what
   * row and col values are given to it, even negative
   * values.  
   * This means that the rest of the system never has to 
   * worry about the bounds of the grid.
   */
  public ICell getCellAt(int row, int col) {
    
    // Remove any additive factors of the size of the grid
    // using the modulo function.
    row %= NCELLROWS;
    col %= NCELLCOLS;
    
    // Turn negative values into positive values.
    if(row<0) row += NCELLROWS;
    if(col<0) col += NCELLCOLS;
    
    return cells[row][col];
  }
  
  /**
   * Maps a given ILambda over all the Cell objects in the cells array.
   * Use a for-each loop.
   * ILambda.apply() is given the Cell object as its input parameter
   * @param lambda The ILambda to map over the cells array.
   */
  private void map(ILambda lambda) {
    // Prob. 2.b.i: STUDENT TO COMPLETE

    
  }

  /**
   * Maps a given ILambda over all the Cell objects in the cells array.
   * Use a regular counted for-loop.
   * ILambda.apply() is given the Cell object, and the current row and column 
   * as its input parameters, in that order.
   * @param lambda The ILambda to map over the cells array.
   */
  private void mapRowCol(ILambda lambda) {
    // Prob. 2.a.ii: STUDENT TO COMPLETE

    
  }
}