package rac;

import listFW.*;
import listFW.factory.*;
import lrs.*;

/**
 * Implements a factory for restricted access containers.  These
 * restricted access containers are implemented using an LRStruct to
 * hold the data objects.
 */
public abstract class ALRSRACFactory2 implements IRACFactory2 {
  private static final IAlgo CheckEmpty = new IAlgo(){
    
    public Object emptyCase(LRStruct host, Object... input) {
      return true;
    }
    
    public Object nonEmptyCase(LRStruct host, Object... input) {
      return false;
    }
  };
  
  protected static final IAlgo defaultInsertAlgo = new IAlgo(){
    
    public Object emptyCase(LRStruct host, Object... input) {
      return host.insertFront(input[0]);
    }
    
    public Object nonEmptyCase(LRStruct host, Object... input) {
      return host.insertFront(input[0]);
    }
  };
  
  
  protected static final IAlgo defaultRemoveAlgo = new IAlgo(){
    
    public Object emptyCase(LRStruct host, Object... nu) {
      throw new IllegalStateException("RAC is empty--nothing to remove!");
    }
    
    public Object nonEmptyCase(LRStruct host, Object... nu) {
      return host;
    }
  };
  /**
   * Implements a general-purpose restricted access container using
   * an LRStruct.  How? 
   *
   * The next item to remove is always at the front of the list of
   * contained objects.  This is invariant!
   *
   * Insertion is, however, delegated to a strategy routine; and
   * this strategy is provided to the container.  This strategy
   * varies to implement the desired kind of container, e.g., queue
   * vs. stack.
   *
   * This inner class is protected so that classes derived from its
   * factory can reuse it to create other kinds of restricted access
   * container. 
   */
  protected class LRSRAContainer implements IRAContainer {
    private IAlgo _insertStrategy;
    private IAlgo _removeStrategy;
    private LRStruct _lrs;
    
    public LRSRAContainer(IAlgo insStrategy, IAlgo remStrategy) {
      _insertStrategy = insStrategy;
      _removeStrategy = remStrategy;
      _lrs = new LRStruct();
    }
    
    /**
     * Empty the container.
     */
    public void clear() {
      _lrs = new LRStruct();
    }
    
    /**
     * Return TRUE if the container is empty; otherwise, return
     * FALSE.
     */
    public boolean isEmpty() {
      return (Boolean)_lrs.execute(CheckEmpty);
    }
    
    /**
     * Return TRUE if the container is full; otherwise, return
     * FALSE. 
     *
     * This implementation can hold an arbitrary number of
     * objects.  Thus, always return false.
     */
    public boolean isFull() {
      return false;
    }
    
    /**
     * Return an immutable list of all elements in the container.
     */
    public IList elements(final IListFactory fact) {
      return (IList)_lrs.execute(new IAlgo() {               
        public Object emptyCase(LRStruct host, Object... nu) {
          return fact.makeEmptyList();
        }
        
        public Object nonEmptyCase(LRStruct host, Object... nu) {
          return fact.makeNEList(host.getFirst(),
                                 (IList)host.getRest().execute(this));
        }
      });
    }
    
    /**
     * Remove the next item from the container and return it.
     */
    public Object get() {
      return ((LRStruct)_lrs.execute(_removeStrategy)).removeFront();
    }
    
    /**
     * Add an item to the container.
     */
    public void put(Object input) {
      _lrs.execute(_insertStrategy, input);
    }
    
    public Object peek() {
      return ((LRStruct)_lrs.execute(_removeStrategy)).getFirst();
    }
  }    
}


